import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn import Parameter


class Encoder(nn.Module):
    """
    Encoder for transcript modality
    """

    def __init__(self, in_dims, hidden_dims):
        super().__init__()
        self.in_features = in_dims
        self.hidden_features = hidden_dims

        self.weights = torch.nn.ParameterList()
        self.layers = len(hidden_dims)

        current_dim = in_dims
        for hidden_dim in hidden_dims:
            weight = Parameter(torch.FloatTensor(current_dim, hidden_dim))
            torch.nn.init.xavier_uniform_(weight)
            self.weights.append(weight)
            current_dim = hidden_dim

    def forward(self, x, adj):
        for i in range(self.layers):
            x = torch.mm(x, self.weights[i])
            x = torch.mm(adj, x)
            # in hidden layers, we use ReLU activation
            if i < self.layers - 1:
                x = F.relu(x)
        return x


class Decoder(nn.Module):
    """
    Decoder for transcript modality
    """

    def __init__(self, hidden_dims, out_dims):
        super().__init__()
        self.hidden_features = hidden_dims
        self.out_features = out_dims

        self.weights = torch.nn.ParameterList()

        current_dim = hidden_dims[0]
        for hidden_dim in hidden_dims[1:] + [out_dims]:
            weight = Parameter(torch.FloatTensor(current_dim, hidden_dim))
            torch.nn.init.xavier_uniform_(weight)
            self.weights.append(weight)
            current_dim = hidden_dim

    def forward(self, x, adj):
        for i, weight in enumerate(self.weights):
            x = torch.mm(x, weight)
            x = torch.mm(adj, x)
            if i < len(self.weights) - 1:
                x = F.relu(x)
        return x


class Projector(nn.Module):
    """
    Projector part, project features of different modality to shared subspace
    """

    def __init__(self, project_dims):
        super().__init__()

        self.layers = nn.ModuleList()
        in_dim = project_dims[0]

        for out_dim in project_dims[1:]:
            self.layers.append(nn.Sequential(
                nn.Linear(in_features=in_dim, out_features=out_dim),
                nn.ELU(),
                nn.Dropout(0.2),
            ))
            in_dim = out_dim

        if len(project_dims) > 1:
            self.layers.append(nn.Sequential(
                nn.Linear(in_features=in_dim, out_features=project_dims[-1]),
                nn.ELU(),
            ))

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        return x


class Discriminator(nn.Module):
    """
    Define the discriminator to distinguish postive sample from negative sample
    """

    def __init__(self, n_h):
        super().__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        """
        Initialize the parameters of the discriminator model
        """
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c, h_pl, h_mi, s_bias1=None, s_bias2=None):
        c_x = c.expand_as(h_pl)
        sc_1 = self.f_k(h_pl, c_x)
        sc_2 = self.f_k(h_mi, c_x)

        if s_bias1 is not None:
            sc_1 = sc_1 + s_bias1
        if s_bias2 is not None:
            sc_2 = sc_2 + s_bias2

        logit = torch.cat((sc_1, sc_2), 1)
        return logit


class AvgReadout(nn.Module):
    """
    We define the readout function here, where the local-avg representation is used to replace the global
    representation in Deep Graph Infomax
    """

    def __init__(self):
        super().__init__()

    def forward(self, emb, mask=None):
        """
        Calculate the avg representation of one micro-environment
        :param emb: the representation of spot
        :param mask: the weighted cell network
        """
        v_sum = torch.mm(mask, emb)
        row_sum = torch.sum(mask, 1)
        row_sum = row_sum.expand((v_sum.shape[1], row_sum.shape[0])).T
        global_emb = v_sum / row_sum
        return F.normalize(global_emb, p=2, dim=1)


class ImageEncoder(nn.Module):
    def __init__(self, in_dims, hidden_dims):
        super().__init__()
        self.in_features = in_dims
        self.hidden_features = hidden_dims
        self.encoder = nn.Sequential(
            nn.Linear(in_features=in_dims, out_features=512),
            nn.BatchNorm1d(512),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(512, hidden_dims),
            nn.BatchNorm1d(hidden_dims),
            nn.ReLU(),
        )

    def forward(self, image_feature):
        image_hidden = self.encoder(image_feature)
        return image_hidden


class ImageDecoder(nn.Module):
    def __init__(self, hidden_dims, in_dims):
        super().__init__()
        self.decoder = nn.Sequential(
            nn.Linear(hidden_dims, 512),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(512, in_dims)
        )

    def forward(self, hidden_features):
        reconstruct_image = self.decoder(hidden_features)
        return reconstruct_image


class ImageAutoEncoder(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.encoder = ImageEncoder(in_dims=in_features, hidden_dims=out_features)
        self.decoder = ImageDecoder(hidden_dims=out_features, in_dims=in_features)

        self.disc = Discriminator(self.out_features)
        self.sigma = nn.Sigmoid()

    def forward(self, image_feat, aug_image_feat):
        z_raw = self.encoder(image_feat)
        z_aug = self.encoder(aug_image_feat)
        recon_img = self.decoder(z_raw)
        return z_raw, z_aug, recon_img


class MuCST(nn.Module):
    def __init__(self, gene_dims, img_dims, graph_nei, project_dims=[64, 64]):
        super().__init__()
        self.graph = graph_nei
        # in_dim -> hidden -> project[0]
        self.gene_encoder_layer1 = Encoder(in_dims=gene_dims[0], hidden_dims=gene_dims[1:])

        self.decoder = Decoder(hidden_dims=list(reversed(gene_dims[1:])), out_dims=gene_dims[0])

        self.img_encoder = ImageEncoder(in_dims=img_dims[0], hidden_dims=img_dims[-1])
        self.img_decoder = ImageDecoder(hidden_dims=img_dims[-1], in_dims=img_dims[0])

        self.disc = Discriminator(gene_dims[-1])
        self.sigma = nn.Sigmoid()
        self.read = AvgReadout()

        self.projector = Projector(project_dims=project_dims)
        self.mse_loss = nn.MSELoss()

    def forward_image(self, image):
        zi = self.img_encoder(image)
        hi = self.projector(zi)
        return zi, hi

    def forward_gene(self, gene, graph):
        zg = self.gene_encoder_layer1.forward(x=gene, adj=graph)
        hg = self.projector(zg)
        return zg, hg

    def recon_gene_loss(self, zg, xg, graph):
        zg = self.decoder(zg, graph)
        return self.mse_loss(zg, xg)

    def recon_img_loss(self, zi, xi):
        zi = self.img_decoder(zi)
        return self.mse_loss(zi, xi)

    def forward(self, gene, fake_gene, img, aug_img1, aug_img2, graph, fusion=False, alpha=1):
        # encode the gene expression into latent embeddings
        zg, hg = self.forward_gene(gene, graph)
        zg_fake, hg_fake = self.forward_gene(fake_gene, graph)

        emb_true = F.relu(zg)
        emb_fake = F.relu(zg_fake)
        g = self.read(emb_true, self.graph)
        g = self.sigma(g)
        g_fake = self.read(emb_fake, self.graph)
        g_fake = self.sigma(g_fake)

        dis_a = self.disc(g, emb_true, emb_fake)

        dis_b = self.disc(g_fake, emb_fake, emb_true)

        zi, hi = self.forward_image(img)
        aug_zi1, aug_hi1 = self.forward_image(aug_img1)
        aug_zi2, aug_hi2 = self.forward_image(aug_img2)

        if fusion:
            z_f = alpha * zg + (1 - alpha) * zi
            rec_gene = self.decoder(z_f, graph)
            rec_img = self.img_decoder(z_f)
        else:
            rec_gene = self.decoder(zg, graph)
            rec_img = self.img_decoder(zi)
        # rec_gene = F.relu(rec_gene)
        return zg, zi, aug_zi1, aug_zi2, hg, hi, aug_hi1, aug_hi2, dis_a, dis_b, rec_gene, rec_img


class OurSimulationModel(nn.Module):
    def __init__(self, gene_dims, img_dims, graph_nei, project_dims=[64, 64]):
        super().__init__()
        self.out_features = gene_dims[1]
        self.graph = graph_nei
        # in_dim -> hidden -> project[0]
        self.gene_encoder_layer1 = Encoder(in_dims=gene_dims[0], hidden_dims=gene_dims[1])
        self.gene_encoder_layer2 = nn.Sequential(
            nn.Linear(in_features=gene_dims[1], out_features=project_dims[0]),
            nn.BatchNorm1d(project_dims[0]),
            nn.ELU(),
            nn.Dropout(0.2)
        )

        self.decoder = Decoder(hidden_dims=project_dims[0], in_dims=gene_dims[0])

        self.img_encoder = ImageEncoder(in_dims=img_dims[0], hidden_dims=project_dims[0])
        self.img_decoder = ImageDecoder(hidden_dims=project_dims[0], in_dims=img_dims[0])

        self.disc = Discriminator(self.out_features)
        self.sigma = nn.Sigmoid()
        self.read = AvgReadout()

        self.projector = nn.Sequential(
            nn.Linear(project_dims[0], project_dims[0]),
            nn.ReLU(),
            nn.Linear(project_dims[0], project_dims[1]),
        )
        self.mse_loss = nn.MSELoss()

    def forward_image(self, image):
        zi = self.img_encoder(image)
        hi = self.projector(zi)
        return zi, hi

    def forward_gene(self, gene, graph):
        zg = self.gene_encoder_layer1.forward(x=gene, adj=graph)
        hg = self.projector(zg)
        return zg, hg

    def recon_gene_loss(self, zg, xg, graph):
        zg = self.decoder(zg, graph)
        return self.mse_loss(zg, xg)

    def recon_img_loss(self, zi, xi):
        zi = self.img_decoder(zi)
        return self.mse_loss(zi, xi)

    def forward(self, gene, fake_gene, img, graph, fusion=False, alpha=1):
        # encode the gene expression into latent embeddings
        zg, hg = self.forward_gene(gene, graph)
        zg_fake, hg_fake = self.forward_gene(fake_gene, graph)

        emb_true = F.relu(zg)
        emb_fake = F.relu(zg_fake)
        g = self.read(emb_true, self.graph)
        g = self.sigma(g)
        g_fake = self.read(emb_fake, self.graph)
        g_fake = self.sigma(g_fake)

        dis_a = self.disc(g, emb_true, emb_fake)

        dis_b = self.disc(g_fake, emb_fake, emb_true)
        zi, hi = self.forward_image(img)

        # reconstruct the gene expression data
        if fusion:
            z_f = alpha * zg + (1 - alpha) * zi
            rec_gene = self.decoder(z_f, graph)
            rec_img = self.img_decoder(z_f)
        else:
            rec_gene = self.decoder(zg, graph)
            rec_img = self.img_decoder(zi)
        return zg, zi, hg, hi, dis_a, dis_b, rec_gene, rec_img



import os
import random

import matplotlib as mpl
import matplotlib.pyplot as plt
import ot
import scanpy as sc
import pandas as pd
import numpy as np
import scipy.sparse as sp
import torch
from matplotlib.collections import LineCollection

from sklearn.linear_model import LinearRegression
from sklearn.decomposition import PCA
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import NearestNeighbors, KDTree, BallTree
from torch.backends import cudnn
from tqdm import tqdm
from scipy.sparse import csr_matrix
import sys
from scipy.optimize import linear_sum_assignment


def read_10x_visium(path, count_file='filtered_feature_bc_matrix.h5', library_id=None, load_images=True,
                    quality='hires', image_path=None):
    """
    Method to load 10X Visium dataset
    :param path: the data path
    :param count_file: the name of count file
    :param library_id: the section id of current slice
    :param load_images: whether load the histology image of slice
    :param quality: select which quality of the histology to load
    :param image_path: set the data path of the image
    :return: the loaded AnnData object
    """
    adata = sc.read_visium(path, count_file=count_file, library_id=library_id, load_images=load_images)
    adata.var_names_make_unique()
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]
    if quality == "fulres":
        image_coord = adata.obsm["spatial"]
        img = plt.imread(image_path, 0)
        adata.uns["spatial"][library_id]["images"]["fulres"] = img
    else:
        scale = adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"]
        image_coord = adata.obsm["spatial"] * scale
    adata.obs["image_col"] = image_coord[:, 0]
    adata.obs["image_row"] = image_coord[:, 1]
    adata.uns["spatial"][library_id]["use_quality"] = quality
    return adata


def refine_label(adata, radius=50, key='label'):
    r"""
    Refine the cluster label according to its neighbors
    """
    n_neigh = radius
    new_type = []
    old_type = adata.obs[key].values
    position = adata.obsm['spatial']
    distance = ot.dist(position, position, metric='euclidean')

    n_cell = distance.shape[0]

    for i in range(n_cell):
        vec = distance[i, :]
        index = vec.argsort()
        neigh_type = []
        for j in range(1, n_neigh + 1):
            neigh_type.append(old_type[index[j]])
        max_type = max(neigh_type, key=neigh_type.count)
        new_type.append(max_type)

    new_type = [str(i) for i in list(new_type)]
    return new_type


def cal_spatial_weight(spatial_data, spatial_k, spatial_type='BallTree'):
    r"""
    Construct the cell network
    :param spatial_data: the coordinates of cells, sucha as adata.obsm['spatial']
    :param spatial_k: the number of neighbors of cell
    :param spatial_type: select which way to construct the cell network
    :return: the constructed cell network
    """
    if spatial_type == 'NearestNeighbors':
        nbrs = NearestNeighbors(n_neighbors=spatial_k + 1, algorithm='ball_tree').fit(spatial_data)
        _, indices = nbrs.n_neighbors(spatial_data)
    elif spatial_type == 'KDTree':
        tree = KDTree(spatial_data, leaf_size=2)
        _, indices = tree.query(spatial_data, k=spatial_k + 1)
    elif spatial_type == 'BallTree':
        tree = BallTree(spatial_data, leaf_size=2)
        _, indices = tree.query(spatial_data, k=spatial_k + 1)
    indices = indices[:, 1:]
    spatial_weight = np.zeros((spatial_data.shape[0], spatial_data.shape[0]))
    for i in range(indices.shape[0]):
        ind = indices[i]
        for j in ind:
            spatial_weight[i][j] = 1
    return spatial_weight


def cal_gene_weight(data, n_components=50, gene_dist_type='cosine'):
    r"""
    Calculate the similarity of gene expression between spots
    :param data: the count matrix of gene expression
    :param n_components: reduce the dimension of expression
    :param gene_dist_type: select the similarity type
    :return: the calculated similarity matrix
    """
    if isinstance(data, csr_matrix):
        data = data.toarray()
    if data.shape[1] > 500:
        pca = PCA(n_components=n_components)
        data = pca.fit_transform(data)
    gene_correlation = 1 - pairwise_distances(data, metric=gene_dist_type)
    return gene_correlation


def find_adjacent_spot(adata, use_data='raw', neighbor_k=6):
    r"""
    for cell i, find its neighbors
    """
    if use_data == 'raw':
        if isinstance(adata.X, csr_matrix):
            gene_matrix = adata.X.toarray()
        elif isinstance(adata.X, np.ndarray):
            gene_matrix = adata.X
        elif isinstance(adata.X, pd.DataFrame):
            gene_matrix = adata.X.values
        else:
            raise ValueError(f"{type(adata.X)} is not a valid type.")
    else:
        gene_matrix = adata.obsm[use_data]

    weights_list = []
    final_coordinates = []
    with tqdm(total=len(adata), desc='Find adjacent spots of each spot',
              bar_format='{l_bar}{bar} [Time left: {remaining}]', ) as pbar:
        for i in range(adata.shape[0]):
            current_spot = adata.obsm['weight_phy_mor'][i].argsort()[-neighbor_k:][: -1]
            spot_weight = adata.obsm['weight_phy_mor'][i][current_spot]
            spot_matrix = gene_matrix[current_spot]
            if spot_weight.sum() > 0:
                spot_weight_scaled = (spot_weight / spot_weight.sum())
                weights_list.append(spot_weight_scaled)
                spot_matrix_scaled = np.multiply(spot_weight_scaled.reshape(-1, 1), spot_matrix)
                spot_matrix_final = np.sum(spot_matrix_scaled, axis=0)
            else:
                spot_matrix_final = np.zeros(gene_matrix.shape[1])
                weights_list.append(np.zeros(len(current_spot)))
            final_coordinates.append(spot_matrix_final)
            pbar.update(1)
    adata.obsm['adjacent_data'] = np.array(final_coordinates)
    adata.obsm['adjacent_weight'] = np.array(weights_list)
    return adata


def data_augmentation(adata, k=6, aug_para=0.3):
    if isinstance(adata.X, csr_matrix):
        gene_matrix = adata.X.A
    elif isinstance(adata.X, np.ndarray):
        gene_matrix = adata.X
    elif isinstance(adata.X, pd.DataFrame):
        gene_matrix = adata.X.values
    else:
        raise ValueError(f"{type(adata.X)} is not a valid type here")

    weights_list = []
    final_coordinates = []

    adj_spatial_mor = adata.obsm['mor_adj']
    for i in range(adata.shape[0]):
        current_spot = adj_spatial_mor[i].argsort()[-k:][: -1]
        spot_weight = adj_spatial_mor[i][current_spot]
        spot_feature_matrix = gene_matrix[current_spot]
        if spot_weight.sum() != 0:
            spot_weight_scaled = spot_weight / spot_weight.sum()
            weights_list.append(spot_weight_scaled)
            spot_matrix_scaled = np.multiply(spot_weight_scaled.reshape(-1, 1), spot_feature_matrix)
            spot_matrix_final = np.sum(spot_matrix_scaled, axis=0)
            final_coordinates.append(spot_matrix_final)
        else:
            print(i, current_spot)
            final_coordinates.append(gene_matrix[i])

    adjacent_spot_data = np.array(final_coordinates)

    if isinstance(adata.X, np.ndarray):
        augment_gene_matrix = adata.X + aug_para * adjacent_spot_data.astype(float)
    elif isinstance(adata.X, csr_matrix):
        augment_gene_matrix = adata.X.A + aug_para * adjacent_spot_data.astype(float)
    else:
        raise ValueError(f"{type(adata.X)} is not a valid type here")

    if aug_para > 0:
        adata.obsm['augment_gene_data'] = augment_gene_matrix
    else:
        adata.obsm['augment_gene_data'] = adata.X


def data_augmentation_all(adata, adjacent_weight=0.3, neighbour_k=4, spatial_k=30, n_components=50, md_dist_type='cosine',
                      gb_dist_type='correlation', use_morphological=True, use_data='raw', spatial_type='KDTree'):
    r"""
    Perform data augmentation
    :param adata: input AnnData object
    :param adjacent_weight: the effect of neighbors to current spot, default is 0.3 according to DeepST
    :param neighbour_k: the number of neighbors of spot i in both spatial proximity and morphological similarity
    :param spatial_k: the number of neighbors of spot i in spatial proximity
    :param n_components: the dimension of gene expression to calculate expression similarity
    :param md_dist_type: the metric of similarity
    :param gb_dist_type: the distance metric of gene expression between spots
    :param use_morphological: whether the morphological information is used
    :param use_data: set which feature is employed to find adjacent spots
    :param spatial_type: the method to construct the KNN graph
    :return: the AnnData object, where the augmented data are stored in 'adata.obsm['augment_gene_data']'
    """
    if use_morphological:
        if spatial_type == 'LinearRegress':
            image_row = adata.obs['image_row']
            image_col = adata.obs['image_col']
            array_row = adata.obs['array_row']
            array_col = adata.obs['array_col']
            rate = 3
            reg_row = LinearRegression().fit(array_row.values.reshape(-1, 1), image_row)
            reg_col = LinearRegression().fit(array_col.values.reshape(-1, 1), image_col)
            # calculate the euc distance between spots
            physical_distance = pairwise_distances(adata.obs[['image_col', 'image_row']], metric='euclidean')
            unit = np.sqrt(reg_row.coef_ ** 2 + reg_col.coef_ ** 2)
            physical_distance = np.where(physical_distance >= rate * unit, 0, 1)
        else:
            physical_distance = cal_spatial_weight(adata.obsm['spatial'], spatial_k=spatial_k,
                                                   spatial_type=spatial_type)
    else:
        physical_distance = cal_spatial_weight(adata.obsm['spatial'], spatial_k=spatial_k, spatial_type=spatial_type)

    print('Physical distance calculating done!')
    print('The number of nearest tie neighbors in physical distance is: {}'.format
          (physical_distance.sum() / adata.shape[0]))

    # calculate gene_expression weight
    gene_correlation = cal_gene_weight(adata[:, adata.var['highly_variable']].X.copy(), gene_dist_type=gb_dist_type,
                                       n_components=n_components)
    gene_correlation[gene_correlation < 0] = 0

    print('Gene correlation calculating done!')
    adata.obsm['gene_correlation'] = gene_correlation
    adata.obsm['physical_distance'] = physical_distance

    # calculate image similarity
    if use_morphological:
        morphological_similarity = 1 - pairwise_distances(np.array(adata.obsm['image_feat_pca']), metric=md_dist_type)
        morphological_similarity[morphological_similarity < 0] = 0
        print('Morphological similarity calculating done!')
        adata.obsm['morphological_similarity'] = morphological_similarity
        # MS_{ij} * GC_{ij} * SW_{ij}
        adata.obsm['weight_matrix_all'] = morphological_similarity * gene_correlation * physical_distance
        adata.obsm['weight_phy_mor'] = morphological_similarity * physical_distance
    else:
        # GC_{ij} * SW_{ij}
        adata.obsm['weight_matrix_all'] = gene_correlation * physical_distance
        adata.obsm['weight_phy_mor'] = physical_distance
    print("The weight result of image feature is added to adata.obsm['weights_matrix_all'].")
    adata = find_adjacent_spot(adata=adata, use_data=use_data, neighbor_k=neighbour_k)
    # augment_gene_data
    if isinstance(adata.X, np.ndarray):
        augment_gene_matrix = adata.X + adjacent_weight * adata.obsm['adjacent_data'].astype(float)
    elif isinstance(adata.X, csr_matrix):
        augment_gene_matrix = adata.X.toarray() + adjacent_weight * adata.obsm['adjacent_data'].astype(float)
    adata.obsm['augment_gene_data'] = augment_gene_matrix
    return adata


def search_res(adata, n_clusters, method='leiden', use_rep='emb', start=0.1, end=3.0, increment=0.01):
    r"""
    Automatically search for a resolution that matches the target number of clusters
    :param adata: the input AnnData object
    :param n_clusters: the target cluster number
    :param method: which graph cluster method is employed, it should be 'leiden' or 'louvain'
    :param start: the resolution where to start search
    :param end: the resolution where to end
    :param increment: the increment value when the resolution is not right
    :return: the best resolution correspond to the target cluster number
    """
    print('Searching resolution...')
    label = 0
    sc.pp.neighbors(adata, n_neighbors=20, use_rep=use_rep)
    for res in sorted(list(np.arange(start, end, increment)), reverse=True):
        if method == 'leiden':
            sc.tl.leiden(adata, resolution=res, random_state=42)
            count_unique = len(pd.DataFrame(adata.obs['leiden']).leiden.unique())
            print('resolution={}, cluster number={}'.format(res, count_unique))
        elif method == 'louvain':
            sc.tl.louvain(adata, resolution=res, random_state=42)
            count_unique = len(pd.DataFrame(adata.obs['louvain']).louvain.unique())
            print('resolution={}, cluster number={}'.format(res, count_unique))
        else:
            raise TypeError(f"Expected `method` to be either `leiden` or `louvain`, found `{str(method)}`.")

        if count_unique == n_clusters:
            label = 1
            break
    assert label == 1, 'Resoultion is not found. Please try bigger range or smaller step!'
    return res


def permutation(feature):
    r"""
    The method to random permute the feature matrix
    :param feature: the raw feature matrix
    :return: the generated feature matrix of the attribute graph
    """
    ids = np.arange(feature.shape[0])
    ids = np.random.permutation(ids)
    feature_permutate = feature[ids]
    return feature_permutate


def get_feature(adata, use_rep=None):
    r"""
    Store the feature in the 'adata.obsm'
    """
    if use_rep is not None:
        target_feature = adata.obsm[use_rep]
    else:
        target_feature = adata.X

    if isinstance(target_feature, sp.csc_matrix) or isinstance(target_feature, sp.csr_matrix):
        target_feature = target_feature.toarray()

    feature_fake = permutation(target_feature)
    adata.obsm['feat'] = target_feature
    adata.obsm['feat_fake'] = feature_fake


def construction_interaction(adata, n_neighbor=3):
    r"""
    Construct the cell spatial network
    :param adata: the input AnnData object
    :param n_neighbor: the number of neighbors
    """
    position = adata.obsm['spatial']
    distance_matrix = pairwise_distances(position)
    n_spot = adata.shape[0]
    adata.obsm['distance_matrix'] = distance_matrix
    interaction = np.zeros((n_spot, n_spot))
    for i in range(n_spot):
        vec = distance_matrix[i, :]
        distance = vec.argsort()
        for t in range(1, n_neighbor + 1):
            y = distance[t]
            interaction[i, y] = 1
    adata.obsm['graph_neigh'] = interaction
    adj = interaction
    adj = adj + adj.T
    adj = np.where(adj > 1, 1, adj)
    adata.obsm['adj'] = adj


def add_contrastive_label(adata):
    r"""
    Generate labels for positive samples and negative samples, which is employed to train discriminator
    """
    n_spot = adata.n_obs
    one_matrix = np.ones([n_spot, 1])
    zero_matrix = np.zeros([n_spot, 1])
    label_CSL = np.concatenate([one_matrix, zero_matrix], axis=1)
    adata.obsm['label_CSL'] = label_CSL


def normalize_adj(adj):
    r"""
    Normalize the adjacency graph
    """
    adj = sp.coo_matrix(adj)
    row_sum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(row_sum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    adj = adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt)
    return adj.toarray()


def preprocess_adj(adj):
    r"""
    Self loop
    """
    adj_normalized = normalize_adj(adj) + np.eye(adj.shape[0])
    return adj_normalized


def fix_seed(seed):
    r"""
    Set seed value, ensure reproduce
    """
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False

    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'


def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='rec_feat_pca', random_seed=2023):
    r"""
    The cluster method, for Visium dataset, Mclust always perform better than graph-based cluster methods
    :param adata: the input AnnData object
    :param num_cluster: the target cluster number
    :param used_obsm: the latent which is used to input the cluster method
    :param random_seed: seed value
    """
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')


def clustering(adata, n_clusters=7, radius=50, method='mclust', start=0.1, end=3.0, increment=0.01, refinement=False):
    r"""
    The cluster method
    :param adata: the input AnnData object
    :param n_clusters: the target cluster number
    :param radius: the neighbor radius to refine the cluster label
    :param method: the cluster method, should be 'mclust', 'leiden', or 'louvain'
    :param start: the start resolution
    :param end: the end resolution
    :param increment: the increment value for search resolution
    :param refinement: whether to refine the cluster labels
    """
    pca = PCA(n_components=20, random_state=42)
    embedding = pca.fit_transform(adata.obsm['rec_feature'].copy())
    adata.obsm['rec_feat_pca'] = embedding

    if method == 'mclust':
        mclust_R(adata, used_obsm='rec_feat_pca', num_cluster=n_clusters)
        adata.obs['domain'] = adata.obs['mclust']
    elif method == 'leiden':
        res = search_res(adata, n_clusters=n_clusters, use_rep='rec_feat_pca', method=method, start=start, end=end,
                         increment=increment)
        sc.tl.leiden(adata, resolution=res, random_state=42)
        adata.obs['domain'] = adata.obs['leiden']
    elif method == 'louvain':
        res = search_res(adata, n_clusters=n_clusters, use_rep='rec_feat_pca', method=method, start=start, end=end,
                         increment=increment)
        sc.tl.louvain(adata, resolution=res, random_state=42)
        adata.obs['domain'] = adata.obs['louvain']

    if refinement:
        new_type = refine_label(adata, radius, key='domain')
        adata.obs['domain'] = new_type


def plot_graph_weights(locations,
                       graph,
                       theta_graph=None,  # azimuthal angles
                       max_weight=1,
                       markersize=1,
                       figsize=(8, 8),
                       title: str = None,
                       flip_yaxis: bool = False,
                       ax=None,
                       ) -> None:
    """
    Visualize weights in a spatial graph,
    heavier weights represented by thicker lines
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    edges, weights, theta = [], [], []

    if theta_graph is not None:
        assert isinstance(theta_graph, csr_matrix)
        assert theta_graph.shape[0] == graph.shape[0]

    for start_node_idx in range(graph.shape[0]):

        ptr_start = graph.indptr[start_node_idx]
        ptr_end = graph.indptr[start_node_idx + 1]

        for ptr in range(ptr_start, ptr_end):
            end_node_idx = graph.indices[ptr]

            # append xs and ys as columns of a numpy array
            edges.append(locations[[start_node_idx, end_node_idx], :])
            weights.append(graph.data[ptr])
            if theta_graph is not None:
                theta.append(theta_graph.data[ptr])

    print(f"Maximum weight: {np.amax(np.array(weights))}\n")
    weights /= np.amax(np.array(weights))

    if theta_graph is not None:
        norm = mpl.colors.Normalize(vmin=min(theta), vmax=max(theta), clip=True)
        mapper = mpl.cm.ScalarMappable(norm=norm, cmap="bwr")
        c = [mapper.to_rgba(t) for t in theta]
    else:
        c = "C0"

    line_segments = LineCollection(
        edges, linewidths=weights * max_weight, linestyle='solid', colors=c, alpha=0.7)
    ax.add_collection(line_segments)

    ax.scatter(locations[:, 0], locations[:, 1], s=markersize, c="gray", alpha=.6, )

    if flip_yaxis:  # 0 on top, typical of image data
        ax.set_ylim(ax.get_ylim()[::-1])

    ax.set_aspect('equal', 'datalim')

    if title is not None:
        ax.set_title(title)


def calculate_overlap(input_data, input_gene_name, used_obs, input_domain_number):
    r"""
    Method to calculate the overlap ratio between marker genes and the identified spatial domain
    :param input_data: the input AnnData object
    :param input_gene_name: the name of makrer gene
    :param used_obs: where the cluster label are stored, such as adata.obs['predict']
    :param input_domain_number: the target domain number, should be inclued in the 'adata.obs[used_obs]'
    :return: the overlap ratio between marker gene and the identified spatial domain
    """
    filtered_adata = input_data[input_data.obs[used_obs] == input_domain_number]
    count_matrix = filtered_adata.X.A
    count_matrix = pd.DataFrame(count_matrix)
    count_matrix.index = filtered_adata.obs.index
    count_matrix.columns = filtered_adata.var.index

    threshold = count_matrix[input_gene_name].min()
    high_expressed_name = input_data[input_data[:, input_gene_name].X > threshold].obs_names
    region_cancer_name = filtered_adata.obs_names

    set_gene_high_expressed = set(high_expressed_name)
    set_cancer_region = set(region_cancer_name)

    overlap = set_gene_high_expressed.intersection(set_cancer_region)
    overlap_ratio = len(overlap) / len(set_cancer_region.union(set_cancer_region))
    return overlap_ratio


def BestMap(L1, L2):
    r"""
    Find the best map for predicted cluster labels to ground truth
    """
    L1 = L1.flatten(order='F').astype(float)
    L2 = L2.flatten(order='F').astype(float)
    if L1.size != L2.size:
        sys.exit('size(L1) must == size(L2)')
    Label1 = np.unique(L1)
    nClass1 = Label1.size
    Label2 = np.unique(L2)
    nClass2 = Label2.size
    nClass = max(nClass1, nClass2)

    # For Hungarian - Label2 are Workers, Label1 are Tasks.
    G = np.zeros([nClass, nClass]).astype(float)
    for i in range(0, nClass2):
        for j in range(0, nClass1):
            G[i, j] = np.sum(np.logical_and(L2 == Label2[i], L1 == Label1[j]))

    _, c = linear_sum_assignment(-G)
    newL2 = np.zeros(L2.shape)
    for i in range(0, nClass2):
        newL2[L2 == Label2[i]] = Label1[c[i]]
    return newL2



import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
from torchtoolbox.tools import mixup_criterion
from .loss import Objective
from sklearn.decomposition import PCA
from .model import MuCST
from .utils import preprocess_adj
from tqdm import tqdm


def train_model(adata, gene_dims, img_dims, proj_dims=[64, 64], lamb1=0.3, lamb2=0.1, lamb3=1, gamma=1, pre_epoch=1500,
                cont_epoch=50, pc_dims=20, patience=30, adj_key='mor_adj', device='cpu'):
    best_loss = float('inf')
    epochs_since_improvement = 0
    features = torch.FloatTensor(adata.obsm['feat'].copy()).to(device)
    features_fake = torch.FloatTensor(adata.obsm['feat_fake'].copy()).to(device)
    label_CSL = torch.FloatTensor(adata.obsm['label_CSL']).to(device)

    image_feature = torch.FloatTensor(adata.obsm['image_feature']).to(device)
    aug_image_feature1 = torch.FloatTensor(adata.obsm['aug_image_feature1']).to(device)
    aug_image_feature2 = torch.FloatTensor(adata.obsm['aug_image_feature2']).to(device)

    adj = preprocess_adj(adata.obsm[adj_key])
    adj = torch.FloatTensor(adj).to(device)

    model = MuCST(gene_dims=gene_dims, img_dims=img_dims, project_dims=proj_dims, graph_nei=adj).to(device)
    model.train()
    loss_CSL = nn.BCEWithLogitsLoss()
    loss_cont = Objective(batch_size=adata.shape[0])

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-4, weight_decay=0.)
    print('Begin to train MuCST...')

    if lamb1 == 0:
        total_epoch = pre_epoch
    else:
        total_epoch = pre_epoch + cont_epoch

    epoch_iter = tqdm(range(total_epoch))
    for epoch in epoch_iter:
        latent_gene, latent_img, latent_aug_img1, latent_aug_img2, high_gene, high_img, high_aug_img1, high_aug_img2, ret, ret_fake, rec_data, rec_img = model.forward(
            features, features_fake, image_feature, aug_image_feature1, aug_image_feature2, adj)
        loss_sl_1 = loss_CSL(ret, label_CSL)
        loss_sl_2 = loss_CSL(ret_fake, label_CSL)
        loss_g2g = loss_sl_1 + loss_sl_2

        loss_feat = F.mse_loss(features, rec_data)
        loss_img = F.mse_loss(image_feature, rec_img)

        loss_recon = loss_feat + lamb1 * loss_img
        if epoch < pre_epoch:
            loss = loss_recon + lamb2 * loss_g2g
            # pre_train_loss.append(loss.item())
        else:
            loss_i2i = loss_cont(high_aug_img1, high_aug_img2)
            lam = np.random.beta(0.2, 0.2)
            loss_i2g = mixup_criterion(loss_cont, high_gene, high_img, high_aug_img1, lam)
            loss = loss_recon + lamb2 * loss_g2g + gamma * loss_i2i + lamb3 * loss_i2g
            # cont_loss.append(loss.item())

        if epoch == pre_epoch:
            best_loss = float("inf")
            print(f"# Epoch {epoch}, loss: {loss.item():.3f}, g2g_loss: {loss_g2g.item():.3f}, gene_recon: {loss_feat.item():.3f}, image_recon: {loss_img.item():.3f}")
            print('Init finished, start contrastive learning part')
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if loss.item() < best_loss:
            best_loss = loss.item()
            epochs_since_improvement = 0
        else:
            epochs_since_improvement = epochs_since_improvement + 1

        if epochs_since_improvement == patience:
            print(f"Training stopped after {epoch + 1} epochs due to no improvement in loss.")
            break

        if epoch < pre_epoch:
            epoch_iter.set_description(
                f"# Epoch {epoch}, loss: {loss.item():.3f}, g2g_loss: {loss_g2g.item():.3f}, gene_recon: {loss_feat.item():.3f}, image_recon: {loss_img.item():.3f}")
        else:
            epoch_iter.set_description(
                f"# Epoch {epoch}, loss: {loss.item():.3f}, g2i_loss: {loss_i2g.item():.3f}, g2g_loss: {loss_g2g.item():.3f}, i2i_loss: {loss_i2i.item():.3f}, gene_recon: {loss_feat.item():.3f}, image_recon: {loss_img.item():.3f}")
    print('Optimization of MuCST finished')

    with torch.no_grad():
        model.eval()
        latent_gene, latent_img, latent_aug_img1, latent_aug_img2, high_gene, high_img, high_aug_img1, high_aug_img2, ret, ret_fake, rec_data, rec_img = model.forward(
            features, features_fake, image_feature, aug_image_feature1, aug_image_feature1, adj)
        latent_gene = latent_gene.detach().cpu().numpy()
        latent_img = latent_img.detach().cpu().numpy()
        rec_data = rec_data.detach().cpu().numpy()
        rec_img = rec_img.detach().cpu().numpy()

    adata.obsm['rec_feature'] = rec_data
    adata.layers['My_ReX'] = rec_data

    alpha = lamb1
    pca = PCA(n_components=pc_dims, random_state=2023)
    rec_feat_pca = pca.fit_transform(rec_data)
    rec_img_pca = pca.fit_transform(rec_img)
    adata.obsm['rec_feat_pca'] = rec_feat_pca
    adata.obsm['rec_img_pca'] = rec_img_pca
    adata.obsm['fusion_pca'] = rec_feat_pca + alpha * rec_img_pca
    adata.obsm['latent_gene'] = latent_gene
    adata.obsm['latent_img'] = latent_img
    # return adata
    torch.cuda.empty_cache()


import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class Objective(nn.Module):
    def __init__(self, batch_size, temperature_f=1.):
        super(Objective, self).__init__()
        self.batch_size = batch_size
        self.temperature_f = temperature_f
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.mask = self.mask_correlated_samples(batch_size)
        self.similarity = nn.CosineSimilarity(dim=2)
        self.criterion = nn.CrossEntropyLoss(reduction='sum')

    def mask_correlated_samples(self, n):
        mask = torch.ones((n, n)).fill_diagonal_(0)
        for i in range(n // 2):
            mask[i, n // 2 + i] = 0
            mask[n // 2 + i, i] = 0
        mask = mask.bool()
        return mask

    def forward(self, h_i, h_j):
        """
        The loss of contrastive learning
        :param h_i:
        :param h_j:
        """
        n = 2 * self.batch_size
        h = torch.cat((h_i, h_j), dim=0)

        sim = torch.matmul(h, h.T) / self.temperature_f
        sim_i_j = torch.diag(sim, self.batch_size)
        sim_j_i = torch.diag(sim, -self.batch_size)

        positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(-1, 1)
        mask = self.mask_correlated_samples(n)
        negative_samples = sim[mask].reshape(n, -1)

        labels = torch.zeros(n).to(positive_samples.device).long()
        logits = torch.cat((positive_samples, negative_samples), dim=1)
        loss = self.criterion(logits, labels)
        loss = loss / n
        return loss


def scale_mse(recon_x, x, alpha=0.):
    return F.mse_loss(alpha * (x - recon_x).mean(dim=1, keepdims=True) + recon_x, x)


def contrast_loss(feat1, feat2, tau=0.1, weight=1.):
    sim_matrix = torch.einsum("ik, jk -> ij", feat1, feat2) / torch.einsum(
        "i, j -> ij", feat1.norm(p=2, dim=1), feat2.norm(p=2, dim=1)
    )
    label = torch.arange(sim_matrix.size(0), device=sim_matrix.device)
    loss = F.cross_entropy(input=sim_matrix / tau, target=label) * weight
    return loss


def trivial_entropy(feat, tau=0.1, weight=1.):
    prob_x = F.softmax(feat / tau, dim=1)
    p = prob_x / prob_x.norm(p=1).sum(0)
    loss = (p * torch.log(p + 1e-8)).sum() * weight
    return loss


def cross_instance_loss(feat1, feat2, tau=0.1, weight=1.):
    sim_matrix = torch.einsum("ik, jk -> ij", feat1, feat2) / torch.einsum(
        "i, j -> ij", feat1.norm(p=2, dim=1), feat2.norm(p=2, dim=1)
    )
    entropy = torch.distributions.Categorical(logits=sim_matrix / tau).entropy().mean() * weight
    return entropy


def kl_loss(feat_x1, feat_x2, tau):
    return torch.nn.KLDivLoss()(
        (feat_x1 / tau).log_softmax(1), (feat_x2 / tau).softmax(1)
    )



import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torchvision.models as models
import torchvision.transforms as transforms
from PIL import Image
from torchtoolbox.transform import Cutout
from tqdm import tqdm
from pathlib import Path
from sklearn.decomposition import PCA


class ImageFeature:
    def __init__(self, adata, pca_components=50, verbose=False, seeds=42):
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.adata = adata
        self.pca_components = pca_components
        self.verbose = verbose
        self.seeds = seeds
        # cnn_type means use which model to extract image features. In this paper, we don't use contrast model to
        # handle image features
        self.cnn_type = 'ResNet50'  # or choose ResNet50

    def load_cnn_model(self):
        """
        load image feature extractor model, we are considering add more pre-trained model
        """
        cnn_pretrained_model = models.resnet50(weights=models.ResNet50_Weights.DEFAULT)
        cnn_pretrained_model.fc = nn.Identity()
        cnn_pretrained_model = cnn_pretrained_model.to(device=self.device)
        return cnn_pretrained_model

    def extract_image_features(self):
        """
        Extract the morphological information from image patches
        :return: the AnnData object, where the morphological features are stored in 'adata.obsm['image_feature']'
        """
        transform_list = [transforms.ToTensor(),
                          transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                               std=[0.229, 0.224, 0.225])]
        aug_transform_list = [
            Cutout(0.5),
            # Convert image from ndarray to tensor
            transforms.ToTensor(),
            # Randomly flip images horizontally
            transforms.RandomHorizontalFlip(p=0.5),
            # Randomly flip images vertically
            transforms.RandomVerticalFlip(p=0.5),
            # Randomly applies a color warp transform
            transforms.RandomApply([transforms.ColorJitter(0.4, 0.4, 0.4, 0.1)], p=0.8),
            # Randomly convert images to grayscale to increase data diversity
            transforms.RandomGrayscale(p=0.2),
            # Normalized the values of image pixel
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ]

        image_to_tensor = transforms.Compose(transform_list)
        aug_image_to_tensor = transforms.Compose(aug_transform_list)
        # load the pretrained cnn model here, and to set to 'eval' model
        model = self.load_cnn_model()
        model.eval()
        feature_df = pd.DataFrame()
        aug_feature_df1 = pd.DataFrame()
        aug_feature_df2 = pd.DataFrame()

        if 'slice_path' not in self.adata.obs.keys():
            raise ValueError('Please run the function image_crop first')

        with tqdm(total=len(self.adata), desc='Extract image feature',
                  bar_format='{l_bar}{bar} [ time left: {remaining} ]', ) as pbar:
            for spot, slice_path in self.adata.obs['slice_path'].items():
                spot_slice = Image.open(slice_path)
                spot_slice = spot_slice.resize((224, 224))
                spot_slice = np.asarray(spot_slice, dtype=np.float32)
                spot_slice = spot_slice / 255.0

                tensor = image_to_tensor(spot_slice)
                tensor = tensor.reshape(1, 3, 224, 224)
                tensor = tensor.to(self.device)
                result = model(tensor)
                result_npy = result.data.cpu().numpy().ravel()
                feature_df[spot] = result_npy
                feature_df = feature_df.copy()

                aug_tensor1 = aug_image_to_tensor(spot_slice)
                aug_tensor1 = aug_tensor1.reshape(1, 3, 224, 224)
                aug_tensor1 = aug_tensor1.to(self.device)
                result = model(aug_tensor1)
                result_npy = result.data.cpu().numpy().ravel()
                aug_feature_df1[spot] = result_npy
                aug_feature_df1 = aug_feature_df1.copy()

                aug_tensor2 = aug_image_to_tensor(spot_slice)
                aug_tensor2 = aug_tensor2.reshape(1, 3, 224, 224)
                aug_tensor2 = aug_tensor2.to(self.device)
                result = model(aug_tensor2)
                result_npy = result.data.cpu().numpy().ravel()
                aug_feature_df2[spot] = result_npy
                aug_feature_df2 = aug_feature_df2.copy()

                pbar.update(1)

        self.adata.obsm['image_feature'] = feature_df.transpose().to_numpy()
        self.adata.obsm['aug_image_feature1'] = aug_feature_df1.transpose().to_numpy()
        self.adata.obsm['aug_image_feature2'] = aug_feature_df2.transpose().to_numpy()
        if self.verbose:
            print("The image feature is added to adata.obsm['Image_Feature']")

        pca = PCA(n_components=self.pca_components, random_state=self.seeds)
        pca.fit(feature_df.transpose().to_numpy())
        self.adata.obsm['image_feat_pca'] = pca.transform(feature_df.transpose().to_numpy())
        if self.verbose:
            print("The pca result of image features is added to adata.obsm['image_feat_pca']")
        return self.adata


def image_crop(adata, save_path, library_id=None, crop_size=112, target_size=224, verbose=False):
    """
    Segment histology image into multiple image patches
    :param adata: input AnnData object
    :param save_path: the path where the image patches will be saved
    :param library_id: section id, typically stored in adata.uns['spatial']
    :param crop_size: the crop size of image patch
    :param target_size: the target image size, for ResNet-50, the value is 224 by default
    :param verbose: control whether show the image crop process
    :return: the AnnData object, where the paths of image patches are stored in 'adata.obs['slice_path']'
    """
    if library_id is None:
        library_id = list(adata.uns['spatial'].keys())[0]
    # load hires quality image
    image = adata.uns['spatial'][library_id]['images']['hires']
    if image.dtype == np.float32 or image.dtype == np.float64:
        image = (image * 255).astype(np.uint8)

    scale = adata.uns["spatial"][library_id]["scalefactors"]["tissue_hires_scalef"]
    image_coord = adata.obsm['spatial'] * scale
    adata.obs['image_col'] = image_coord[:, 0]
    adata.obs['image_row'] = image_coord[:, 1]
    image_pillow = Image.fromarray(image)
    tile_names = []

    with tqdm(total=len(adata), desc='Tiling Image', bar_format='{l_bar}{bar} [ time left: {remaining} ]') as pbar:
        for image_row, image_col in zip(adata.obs['image_row'], adata.obs['image_col']):
            image_down = image_row - crop_size / 2
            image_up = image_row + crop_size / 2
            image_left = image_col - crop_size / 2
            image_right = image_col + crop_size / 2

            tile = image_pillow.crop(
                (image_left, image_down, image_right, image_up)
            )
            tile.thumbnail((target_size, target_size), Image.LANCZOS)
            tile.resize((target_size, target_size))
            tile_name = str(image_col) + '-' + str(image_row) + '-' + str(crop_size)
            if save_path is not None:
                out_tile = Path(save_path) / (tile_name + '.png')
                tile_names.append(str(out_tile))
                if verbose:
                    print('Generating tile at location ({}, {})'.format(str(image_col), str(image_row)))
                tile.save(out_tile, 'PNG')
            pbar.update(1)

    adata.obs['slice_path'] = tile_names
    if verbose:
        print("The slice path of image feature is added to adata.obs['slice_path']")
    return adata
